#ifndef CONFIG_H
#define CONFIG_H

#include <limits.h>
#include <stdbool.h>
#include <X11/Xlib.h>
#include <X11/X.h>
#include <sys/types.h>

#define CONFIG_COMMAND_MAX_LEN 128
#define CONFIG_PARSERRESULT_MAX 1024
#define CONFIG_ARGC_MAX 16
#define CONFIG_ARGLEN_MAX 256
#define CONFIG_COLOR_MAX 16

#define PARSER_RETURN(result, _success, fmt, ...)        \
    do { (result).success = (_success);                  \
    snprintf((result).message, CONFIG_PARSERRESULT_MAX, \
    (fmt), ##__VA_ARGS__); return (result);} while (0);

typedef struct Client Client;
typedef struct Wm Wm;
typedef struct Arg Arg;
typedef struct Keybind Keybind;

struct Arg {
    int i;
    unsigned int ui;
    float f;
    Client *c;
    char* s;
    char **sl;

    int count;
};

typedef void (*KeybindFunction)(Wm*, Arg*) ;

struct Keybind {
    unsigned int mask;
    KeySym keysym;
    KeybindFunction function;
    Arg args;
};

typedef struct {
    const char command[CONFIG_ARGLEN_MAX];
    const KeybindFunction function;
} KeybindFunctionCommand;

typedef enum {
    TYPE_INTEGER,
    TYPE_UNSIGNED_INTEGER,
    TYPE_BOOL,
    TYPE_STRING,
    TYPE_KEYBIND,
    TYPE_FLOAT,
} CommandArgType;

typedef struct {
    bool success;
    char message[CONFIG_PARSERRESULT_MAX];
} ParserResult;

typedef struct {
    char string[CONFIG_COMMAND_MAX_LEN];
    size_t arg_count;
    off_t cfg_var_offset[CONFIG_ARGC_MAX];
    CommandArgType arg_type;
    size_t argc;
    char **argv;
    Arg *args;
    Keybind keybind;
} ConfigCommand;

typedef struct {
    char file_path[PATH_MAX];
    ConfigCommand *available_commands;
    size_t command_count;
} ConfigFile;

typedef struct {
    float ms_p;
    char border_col[CONFIG_COLOR_MAX];
    char focused_border_col[CONFIG_COLOR_MAX];
    unsigned int border_width;
    Keybind *keybinds;
    int kb_count;
    bool focus_on_motion;
    bool show_titlebar;
    char titlebar_color[CONFIG_COLOR_MAX];
    char titlebar_focused_color[CONFIG_COLOR_MAX];
    char titlebar_font[CONFIG_ARGLEN_MAX];
    char titlebar_text_color[CONFIG_COLOR_MAX];
    char log_path[PATH_MAX];
} Config;

void wm_cfg_init_def(Config *config);
void wm_keybinds_init_def(Config *config);
void wm_keybinds_free(Config *config);
void wm_configfile_init(ConfigFile *config, const char *path);
void wm_configfile_free(ConfigFile *config);
void wm_configfile_read(ConfigFile *cfile, Config *config);
void wm_config_apply_command(Config *config, ConfigCommand *command);
ParserResult wm_configfile_parse_line(const ConfigFile *cfile, Config *config,
                                      char* line, ConfigCommand *command);
void wm_configcommand_init(ConfigCommand *command);
void wm_configcommand_copy(ConfigCommand *dest, const ConfigCommand *src);
void wm_configcommand_argv_parse(ConfigCommand *command);
void wm_configcommand_free(ConfigCommand *command);
bool atobool(char *arr);

Arg wm_keybind_argv_to_args(Keybind kb, ConfigCommand command);

int wm_string_to_direction(const char *str);

KeybindFunction wm_string_to_keybindfunction(const char* str);

/* Converts string to Key mask (see X.h:218-228). */
unsigned int wm_string_to_mask(const char* str);

/* The first element is the mask, the second element is the
 * key. The returned array needs to be freed. */
KeySym* wm_keybind_string_to_mask_keysym(const char* kb_argv);

Keybind wm_config_parse_bind_command(ConfigCommand *command);

/* Finds a keybind matching the function of the keybind parameter, and
 * replaces it. */
void wm_config_replace_keybind(Config *config, const Keybind *keybind);

#endif // CONFIG_H
